from typing import Optional

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.utils import RenderContext, ColorType


class Label(BaseElement):
    def __init__(
            self, rect: pg.Rect, text: str, text_color: ColorType = (200, 200, 200),
            bg_color: Optional[ColorType] = None
    ):
        super().__init__(rect)
        self._text = text
        self.text_color = text_color
        self.bg_color = bg_color
        self._text_surface: Optional[pg.Surface] = None

    def handle_event(self, event: pg.event.Event):
        super().handle_event(event)

    def set_text(self, text: str):
        self._text = text
        self._text_surface = None

    def draw(self, screen: pg.Surface, render_context: RenderContext):
        if self.bg_color:
            pg.draw.rect(screen, self.bg_color, self.rect)

        if self._text:
            if self._text_surface is None:
                self._text_surface = render_context.font.render(self._text, True, self.text_color)
                self.rect = self._text_surface.get_rect(center=self.rect.center)
            screen.blit(self._text_surface, self.rect)



from typing import Iterable, Optional, List

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.utils import RenderContext


class Container:
    def __init__(self):
        self._element_cache: Optional[List[BaseElement]] = None

    def iter_elements(self) -> Iterable[BaseElement]:
        """
        Iter over all elements in the container.
        :return: Iterable of BaseElement objects.
        """
        if self._element_cache is None:
            self._element_cache = [elem for elem in self.__dict__.values() if isinstance(elem, BaseElement)]

        yield from self._element_cache

    def handle_event(self, event: pg.event.Event):
        for elem in self.iter_elements():
            elem.handle_event(event)

    def render(self, screen: pg.Surface, render_context: RenderContext):
        for element in self.iter_elements():
            element.render(screen, render_context)



from typing import Optional, Tuple, List, Union, Iterable

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.utils import RenderContext, ColorType, load_font

SCRAP_TEXT = 'text/plain;charset=utf-8'


'''
Consider the following paragraphs
paragraphs = [
    "123", "456"
]
cursor at the start of the first paragraph _123: Cursor(x, 0, 0)
cursor at the start of the first paragraph 123_: Cursor(x, 3, 0)
the length of a paragraph is the string length + 1.
'''
class Cursor:
    def __init__(self, line_index: int, paragraph_index: int, char_index: int):
        self.line_index = line_index
        self.paragraph_index = paragraph_index
        self.char_index = char_index

    @staticmethod
    def from_tuple(paragraph_index: int, char_index: int):
        return Cursor(-1, paragraph_index, char_index)

    def copy(self) -> 'Cursor':
        return Cursor(self.line_index, self.paragraph_index, self.char_index)

    def __repr__(self):
        return f'Cursor({self.line_index}, {self.paragraph_index}, {self.char_index})'

    def __eq__(self, other):
        return self.line_index == other.line_index and self.paragraph_index == other.paragraph_index and self.char_index == other.char_index

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if self.line_index == other.line_index:
            if self.paragraph_index == other.paragraph_index:
                return self.char_index < other.char_index
            else:
                return self.paragraph_index < other.paragraph_index
        else:
            return self.line_index < other.line_index

    def __gt__(self, other):
        return self.__ne__(other) and not self.__lt__(other)


CursorOrTuple = Union[Cursor, Tuple[int, int]]


class Line:
    """
    Represents a line in a text field. Can contain multiple paragraphs that are automatically wrapped.
    Each manual enter from the user will create a new Line object. Each line that is longer than the width of the
    TextField will create a new paragraph.
    """
    def __init__(self, text: Union[str, List[str]] = ''):
        if isinstance(text, str):
            text = [text]
        self.paragraphs: List[str] = text

    def __repr__(self):
        # return f'Line({len(self.paragraphs)} paragraphs, {self.num_chars()} chars)'
        return str(self.paragraphs)

    def get_line_char_index(self, paragraph_index: int, char_index: int) -> int:
        """
        Returns the character index, if all paragraphs would be in a single string.
        """
        return sum(len(p) + 1 for p in self.paragraphs[:paragraph_index]) + char_index

    def auto_wrap_and_norm_cursor(self, font: pg.font.Font, max_width: int, cursor: Cursor):
        """
        Wraps the line to fit within the given max width.

        :param font: The font to use for wrapping.
        :param max_width: The maximum width to wrap the line to.
        :param cursor: A cursor object, that will be handled correctly, if wrapping occurs at the cursor position. It is
            assumed, that the cursor points to the current line.
        :returns: The given cursor. If a wrap happens the returned cursor will point to the same character possibly in
            a different paragraph.
        """
        line_char_index = self.get_line_char_index(cursor.paragraph_index, cursor.char_index)
        self.auto_wrap(font, max_width)
        new_cursor = Cursor(cursor.line_index, cursor.paragraph_index, line_char_index)
        new_cursor.paragraph_index, new_cursor.char_index = self.get_paragraph_char_index(line_char_index)
        return new_cursor

    def get_paragraph_char_index(self, line_char_index: int) -> Tuple[int, int]:
        cum_char_sum = 0
        for i, p in enumerate(self.paragraphs):
            p_chars = len(p) + 1  # every paragraph ends with a virtual extra character
            if line_char_index < cum_char_sum + p_chars:
                return i, line_char_index - cum_char_sum
            cum_char_sum += p_chars  # TODO: +1?
        return len(self.paragraphs) - 1, 0

    def word_list(self) -> List[str]:
        word_list = []
        for p in self.paragraphs:
            word_list.extend(p.split(' '))
        return word_list

    def auto_wrap(self, font: pg.font.Font, max_width: int):
        words = self.word_list()
        current_line = ''
        new_paragraphs = []

        for word in words:
            # Test if adding this word would exceed max width
            test_line = current_line + (" " if current_line else "") + word
            test_width = font.size(test_line)[0]

            if test_width <= max_width:
                # Word fits on current line
                current_line = test_line
            else:
                # Word doesn't fit
                if current_line:
                    # Save current line and start new one
                    new_paragraphs.append(current_line)
                    current_line = word
                else:
                    # Single word is too long, force it on its own line
                    new_paragraphs.append(word)
                    current_line = ""

        # Add the last line of this paragraph
        if current_line:
            new_paragraphs.append(current_line)

        self.paragraphs = new_paragraphs
        self.ensure_paragraph()

    def num_paragraphs(self) -> int:
        return len(self.paragraphs)

    def num_chars(self) -> int:
        """
        Returns the number of characters in the line. Each automatic wrap counts as one character.
        """
        return sum(len(p) for p in self.paragraphs) + len(self.paragraphs) - 1

    def split(self, paragraph_index: int, split_index: int) -> Tuple['Line', 'Line']:
        """
        Splits the current line at the given paragraph and character index. Returns the two resulting lines.
        """
        left_paragraphs = self.paragraphs[:paragraph_index]
        right_paragraphs = self.paragraphs[paragraph_index+1:]
        middle_paragraph = self.paragraphs[paragraph_index]
        left_part = ' ' + middle_paragraph[:split_index]
        right_part = middle_paragraph[split_index:] + ' '
        left_paragraphs.append(left_part)
        right_paragraphs.insert(0, right_part)
        return Line(''.join(left_paragraphs)), Line(''.join(right_paragraphs))

    @staticmethod
    def _remove_from_paragraph(paragraph: str, start: int, end: int) -> str:
        return paragraph[:start] + paragraph[end:]

    def delete(self, start: CursorOrTuple, end: CursorOrTuple):
        if isinstance(start, tuple):
            start = Cursor.from_tuple(*start)
        if isinstance(end, tuple):
            end = Cursor.from_tuple(*end)

        if start.paragraph_index == end.paragraph_index:
            p = self.paragraphs[start.paragraph_index]
            self.paragraphs[start.paragraph_index] = p[:start.char_index] + p[end.char_index:]
            self.ensure_paragraph()
            return
        before_start = self.paragraphs[:start.paragraph_index]
        last_after_start = self.paragraphs[start.paragraph_index]
        first_before_end = self.paragraphs[end.paragraph_index]
        after_end = []
        if end.paragraph_index+1 < len(self.paragraphs):
            after_end = self.paragraphs[end.paragraph_index+1:]

        new_paragraphs = before_start
        middle_paragraph = last_after_start[:start.char_index] + first_before_end[end.char_index:]
        if middle_paragraph:
            new_paragraphs.append(middle_paragraph)
        new_paragraphs.extend(after_end)
        self.paragraphs = new_paragraphs
        self.ensure_paragraph()

    def ensure_paragraph(self):
        if not self.paragraphs:
            self.paragraphs = ['']

    def end(self) -> Tuple[int, int]:
        paragraph_index = len(self.paragraphs) - 1
        char_index = len(self.paragraphs[-1])
        return paragraph_index, char_index


class TextField(BaseElement):
    def __init__(
            self, rect: pg.Rect, text: str = "", placeholder: str = "",
            bg_color: ColorType = (40, 40, 40), hover_color: ColorType = (50, 50, 50),
            clicked_color: ColorType = (60, 60, 60), border_color: ColorType = (120, 120, 120),
            text_color: ColorType = (200, 200, 200)
    ):
        super().__init__(rect)
        self.lines: List[Line] = []
        self.placeholder = placeholder

        self.bg_color = bg_color
        self.hover_color = hover_color
        self.clicked_color = clicked_color
        self.border_color = border_color
        self.text_color = text_color
        self.border_width = 2
        self.padding = 5

        self.cursor: Cursor = self.end_cursor()
        self.selection_start: Optional[Cursor] = None  # Start of text selection
        self.is_focused: bool = False  # Whether the field is focused
        self.mouse_down: bool = False  # For tracking drag selection
        self.scroll_offset: int = 0  # Vertical scroll offset (in lines)

        self.font = load_font()
        self.line_height = self.font.get_height() if self.font else 20

        self.set_text(text)

    def set_text(self, text: str):
        self.lines = [Line(l) for l in text.split('\n')]
        self._wrap_text()

    def get_text(self) -> str:
        paragraphs = []
        for line in self.lines:
            paragraphs.append(' '.join(p for p in line.paragraphs))
        return '\n'.join(paragraphs)

    def end_cursor(self) -> Cursor:
        if not self.lines:
            return Cursor(0, 0, 0)
        line = self.lines[-1]
        line_index = len(self.lines) - 1
        if not line.paragraphs:
            return Cursor(line_index, 0, 0)
        paragraph = line.paragraphs[-1]
        paragraph_index = len(line.paragraphs)-1
        return Cursor(line_index, paragraph_index, len(paragraph))

    def _wrap_text(self):
        """
        Wrap text to fit within max_width, breaking at word boundaries.
        """
        max_width = self.rect.width - 2 * self.padding
        for line_index, line in enumerate(self.lines):
            if line_index == self.cursor.line_index:
                self.cursor = line.auto_wrap_and_norm_cursor(self.font, max_width, self.cursor)
            else:
                line.auto_wrap(self.font, max_width)

    def _cursor_from_mouse_pos(self, mouse_pos: Tuple[int, int]) -> Optional[Cursor]:
        """Calculate the character index in text based on mouse position."""
        if not self.rect.collidepoint(*mouse_pos):
            return self.end_cursor()

        if not self.lines:
            return Cursor(0, 0, 0)

        rel_x = mouse_pos[0] - self.rect.x - self.padding
        rel_y = mouse_pos[1] - self.rect.y - self.padding

        # Calculate line index based on Y position and scroll
        line_par = self._get_line_and_paragraph_by_y(rel_y)
        if line_par is None:
            return self.end_cursor()
        line_index, paragraph_index = line_par

        line = self.lines[line_index]
        paragraph = line.paragraphs[paragraph_index]

        # Find which paragraph the click is in based on X position
        for num_chars in range(len(paragraph) + 1):
            text = paragraph[:num_chars]
            paragraph_width = self.font.size(text)[0]
            if paragraph_width > rel_x:
                # TODO: num_chars + 1?
                return Cursor(line_index, paragraph_index, num_chars)

        # If click is past all paragraphs, return end of last paragraph
        return Cursor(line_index, paragraph_index, len(paragraph))

    def _get_view_line_pos(self, cursor: Cursor) -> int:
        view_line_pos = 0
        for line in self.lines[:cursor.line_index]:
            view_line_pos += line.num_paragraphs()
        return view_line_pos + cursor.paragraph_index

    def _get_num_paragraphs(self) -> int:
        return sum(l.num_paragraphs() for l in self.lines)

    def _get_line_and_paragraph_by_y(self, ypos: int) -> Optional[Tuple[int, int]]:
        view_line_index = self.scroll_offset + ypos // self.line_height

        start_line_index = 0
        for line_index, line in enumerate(self.lines):
            end_line_index = start_line_index + line.num_paragraphs()
            if view_line_index < end_line_index:
                paragraph_index = view_line_index - start_line_index
                return line_index, paragraph_index
            start_line_index = end_line_index

        # out of region
        return None

    def handle_event(self, event: pg.event.Event):
        super().handle_event(event)

        # Handle mouse button down - start selection
        if event.type == pg.MOUSEBUTTONDOWN and event.button == 1:
            if self.is_hovered:
                self.is_focused = True
                cursor = self._cursor_from_mouse_pos(event.pos)
                if cursor is not None:
                    self.cursor = cursor
                self.selection_start = self.cursor
                self.mouse_down = True
            else:
                self.is_focused = False
                self.selection_start = None

        # Handle mouse drag - update selection
        if event.type == pg.MOUSEMOTION:
            if self.mouse_down and pg.mouse.get_pressed()[0]:
                if self.is_hovered or self.is_focused:
                    cursor = self._cursor_from_mouse_pos(event.pos)
                    if cursor is not None:
                        self.cursor = cursor

        # Handle mouse button up - finish selection
        if event.type == pg.MOUSEBUTTONUP and event.button == 1:
            if self.mouse_down:
                if self.selection_start == self.cursor:
                    self.selection_start = None
            self.mouse_down = False

        # Handle scroll wheel
        if event.type == pg.MOUSEWHEEL and self.is_hovered:
            self.scroll_offset = max(0, self.scroll_offset - event.y)
            self._clamp_scroll()

        # Handle keyboard input only if focused
        if self.is_focused:
            if event.type == pg.KEYDOWN:
                ctrl_pressed = bool(event.mod & pg.KMOD_CTRL)
                shift_pressed = bool(event.mod & pg.KMOD_SHIFT)

                if event.key == pg.K_RETURN:
                    self._create_newline()
                elif event.key == pg.K_BACKSPACE:
                    self._delete_direction(-1, ctrl_pressed)
                elif event.key == pg.K_DELETE:
                    self._delete_direction(1, ctrl_pressed)
                elif event.key == pg.K_LEFT:
                    self._move_my_cursor(-1, ctrl_pressed, shift_pressed)
                elif event.key == pg.K_RIGHT:
                    self._move_my_cursor(1, ctrl_pressed, shift_pressed)
                elif event.key == pg.K_UP:
                    self._move_my_cursor_vertical(-1, shift_pressed)
                elif event.key == pg.K_DOWN:
                    self._move_my_cursor_vertical(1, shift_pressed)
                elif event.key == pg.K_HOME:
                    self._move_cursor_home(shift_pressed, ctrl_pressed)
                elif event.key == pg.K_END:
                    self._move_cursor_end(shift_pressed, ctrl_pressed)
                elif event.key == pg.K_a and ctrl_pressed:
                    self._select_all()
                elif event.key == pg.K_c and ctrl_pressed:
                    self._copy_to_clipboard()
                elif event.key == pg.K_v and ctrl_pressed:
                    self._paste_from_clipboard()
                elif event.key == pg.K_x and ctrl_pressed:
                    self._cut_to_clipboard()
                elif event.unicode and event.unicode.isprintable():
                    self._insert_text(event.unicode)

    def _get_line(self, cursor: Cursor) -> Line:
        return self.lines[cursor.line_index]

    def _get_paragraph(self, cursor: Union[Cursor, Tuple[int, int]]) -> str:
        if isinstance(cursor, Cursor):
            return self._get_line(cursor).paragraphs[cursor.paragraph_index]
        elif isinstance(cursor, tuple):
            return self.lines[cursor[0]].paragraphs[cursor[1]]
        else:
            raise ValueError(f"Invalid cursor type: {type(cursor)}")

    def _create_newline(self):
        """Split a line into two."""
        line_index = self.cursor.line_index
        orig_line = self.lines[line_index]
        left_line, right_line = orig_line.split(self.cursor.paragraph_index, self.cursor.char_index)

        max_width = self.rect.width - 2 * self.padding
        left_line.auto_wrap(self.font, max_width)
        right_line.auto_wrap(self.font, max_width)

        self.lines[line_index] = left_line
        self.lines.insert(line_index + 1, right_line)

        self.cursor = Cursor(line_index + 1, 0, 0)

    @staticmethod
    def _next_char_index(paragraph: str, char_index: int, direction: int, jump_words: bool) -> int:
        """Return the index of the next character in the given paragraph."""
        if not jump_words:
            return char_index + direction

        if direction > 0:
            if char_index == len(paragraph):
                return len(paragraph) + 1
            new_index = paragraph.find(' ', char_index+1)
            if new_index == -1:
                return len(paragraph)
            return new_index
        elif direction < 0:
            if char_index == 0:
                return -1
            new_index = paragraph.rfind(' ', 0, char_index)
            if new_index == -1:
                return 0
            return new_index
        else:
            raise ValueError("Invalid direction")

    def _move_my_cursor(self, direction: int, jump_words: bool = False, select: bool = False):
        """Move the cursor in the direction given by the given direction."""
        self._save_selection_start(select)
        self._move_cursor(self.cursor, direction, jump_words)
        self._update_scroll()

    def _move_cursor(self, cursor: Cursor, direction: int, jump_words: bool = False):
        """
        Modifies the given cursor moving in the given direction (-1 or 1).
        If jump_words is True, the cursor will jump over words.
        """
        wrap_back = False
        wrap_forward = False
        new_line_index = cursor.line_index
        new_paragraph_index = cursor.paragraph_index
        new_char_index = TextField._next_char_index(
            self._get_paragraph(cursor), cursor.char_index, direction, jump_words
        )
        if new_char_index < 0 or new_char_index > len(self._get_paragraph(cursor)):
            wrap_back = new_char_index < 0
            wrap_forward = new_char_index > len(self._get_paragraph(cursor))
            new_paragraph_index = cursor.paragraph_index + direction
            current_line = self._get_line(cursor)
            if new_paragraph_index < 0 or new_paragraph_index >= current_line.num_paragraphs():
                new_line_index = cursor.line_index + direction
                if new_line_index < 0 or new_line_index >= len(self.lines):
                    return
                if new_paragraph_index < 0:
                    new_paragraph_index = self.lines[new_line_index].num_paragraphs() - 1
                elif new_paragraph_index >= current_line.num_paragraphs():
                    new_paragraph_index = 0
        if wrap_forward:
            new_char_index = 0
        if wrap_back:
            new_char_index = len(self._get_paragraph((new_line_index, new_paragraph_index)))

        cursor.line_index = new_line_index
        cursor.paragraph_index = new_paragraph_index
        cursor.char_index = new_char_index

    def _update_scroll(self):
        """Update scroll offset to keep cursor visible."""
        visible_height = self.rect.height - 2 * self.padding
        visible_lines = max(1, int(visible_height / self.line_height))

        cursor_line = self._get_view_line_pos(self.cursor)

        # Scroll down if cursor is below visible area
        if cursor_line >= self.scroll_offset + visible_lines:
            self.scroll_offset = cursor_line - visible_lines + 1

        # Scroll up if cursor is above visible area
        if cursor_line < self.scroll_offset:
            self.scroll_offset = cursor_line

        self._clamp_scroll()

    def _clamp_scroll(self):
        """Ensure scroll offset is within valid bounds."""
        visible_height = self.rect.height - 2 * self.padding
        visible_lines = max(1, int(visible_height / self.line_height))

        max_scroll = max(0, self._get_num_paragraphs() - visible_lines)
        self.scroll_offset = max(0, min(self.scroll_offset, max_scroll))

    def _insert_text(self, char: str):
        """Insert a character at the cursor position."""
        self._delete_selection()

        # insert text
        paragraph = self._get_paragraph(self.cursor)
        paragraph = paragraph[:self.cursor.char_index] + char + paragraph[self.cursor.char_index:]
        line = self.lines[self.cursor.line_index]
        line.paragraphs[self.cursor.paragraph_index] = paragraph

        # wrap line
        self._wrap_text()

        # handle cursor
        self.cursor.char_index += len(char)
        self.selection_start = None
        self._update_scroll()

    def _move_my_cursor_vertical(self, direction: int, select: bool = False):
        """Move the cursor vertically in the direction given by the given direction."""
        self._save_selection_start(select)
        self.cursor = self._move_cursor_vertical(self.cursor, direction)
        self._update_scroll()

    def _move_cursor_vertical(self, cursor: Cursor, direction: int) -> Cursor:
        new_paragraph_index = cursor.paragraph_index + direction
        new_line_index = cursor.line_index
        new_char_index = cursor.char_index

        if new_paragraph_index < 0 or new_paragraph_index >= self._get_line(cursor).num_paragraphs():
            new_line_index += direction
            if new_line_index < 0:
                new_line_index = 0
                new_paragraph_index = 0
                new_char_index = 0
            elif new_line_index >= len(self.lines):
                new_line_index = len(self.lines) - 1
                new_paragraph_index = self.lines[new_line_index].num_paragraphs() - 1
                new_char_index = len(self._get_paragraph((new_line_index, new_paragraph_index)))
            else:
                if new_paragraph_index < 0:
                    new_paragraph_index = self.lines[new_line_index].num_paragraphs() - 1
                elif new_paragraph_index >= self._get_line(cursor).num_paragraphs():
                    new_paragraph_index = 0

        return self._clamp_cursor(Cursor(new_line_index, new_paragraph_index, new_char_index))


    def _move_cursor_home(self, select: bool, ctrl_pressed: bool):
        """Move cursor to start of line or start of text."""
        self._save_selection_start(select)

        if ctrl_pressed:
            # Move to start of text
            self.cursor = Cursor(0, 0, 0)
        else:
            # Move to start of current wrapped line
            self.cursor.char_index = 0

        self._update_scroll()

    def _save_selection_start(self, select: bool):
        if select:
            if self.selection_start is None:
                self.selection_start = self.cursor.copy()
        else:
            self.selection_start = None

    def _move_cursor_end(self, select: bool, ctrl_pressed: bool):
        """Move cursor to end of line or end of text."""
        self._save_selection_start(select)

        if ctrl_pressed:
            # Move to end of text
            self.cursor = self.end_cursor()
        else:
            # Move to end of current wrapped line
            self.cursor.char_index = len(self._get_paragraph(self.cursor))

        self._update_scroll()

    def _clamp_cursor(self, cursor: Cursor) -> Cursor:
        """Ensure cursor is within valid bounds."""
        new_cursor = cursor.copy()
        if new_cursor.line_index >= len(self.lines):
            new_cursor.line_index = len(self.lines) - 1
            new_cursor.paragraph_index = max(0, self.lines[new_cursor.line_index].num_paragraphs() - 1)
            new_cursor.char_index = len(self._get_paragraph(new_cursor))
        elif new_cursor.paragraph_index >= self.lines[new_cursor.line_index].num_paragraphs():
            new_cursor.paragraph_index = self.lines[new_cursor.line_index].num_paragraphs() - 1
            new_cursor.char_index = len(self._get_paragraph(new_cursor))
        elif new_cursor.char_index > len(self._get_paragraph(new_cursor)):
            new_cursor.char_index = len(self._get_paragraph(new_cursor))

        return new_cursor

    def _delete_direction(self, direction: int, jump_words: bool = False):
        """Remove a character in the given direction."""
        if self.selection_start is not None:
            self._delete_selection()
        else:
            target_cursor = self.cursor.copy()
            self._move_cursor(target_cursor, direction, jump_words)
            self._delete(self.cursor, target_cursor)
            self.cursor = min(target_cursor, self.cursor)

    def _get_selection_range(self) -> Tuple[Cursor, Cursor]:
        """Get the start and end of the current selection (ordered)."""
        if self.selection_start is None:
            return self.cursor, self.cursor
        return min(self.selection_start, self.cursor), max(self.selection_start, self.cursor)

    def _delete_selection(self) -> bool:
        """Delete selected text if any. Returns True if text was deleted."""
        start, end = self._get_selection_range()
        if start != end:
            self._delete(start, end)
            self.cursor = start
            self.selection_start = None
            return True
        return False

    def _delete(self, start: Cursor, end: Cursor):
        """Delete text between both cursors."""
        if end < start:
            start, end = end, start
        max_width = self.rect.width - 2 * self.padding

        if start.line_index == end.line_index:
            line = self._get_line(start)
            line.delete(start, end)
            line.auto_wrap(self.font, max_width)
        else:
            new_lines = self.lines[:start.line_index].copy()

            last_start_line = self.lines[start.line_index]
            last_start_line.delete(start, last_start_line.end())

            first_end_line = self.lines[end.line_index]
            first_end_line.delete((0, 0), end)

            middle_line = Line(last_start_line.paragraphs + first_end_line.paragraphs)
            middle_line.ensure_paragraph()
            middle_line.auto_wrap(self.font, max_width)
            new_lines.append(middle_line)

            new_lines.extend(self.lines[end.line_index + 1:])
            self.lines = new_lines
        self.cursor = self._clamp_cursor(start)

    def _select_all(self):
        """Select all text."""
        self.selection_start = Cursor(0, 0, 0)
        self.cursor = self.end_cursor()

    def _copy_to_clipboard(self):
        """Copy selected text to clipboard."""
        start, end = self._get_selection_range()
        if start != end:
            text = '\n'.join(self.iter_paragraphs(start, end))
            pg.scrap.put(SCRAP_TEXT, text.encode('utf-8'))

    def iter_paragraphs(self, start: Cursor, stop: Cursor) -> Iterable[str]:
        """Iterate over paragraphs between start and stop cursors."""
        for line_index in range(start.line_index, stop.line_index + 1):
            paragraphs = self.lines[line_index].paragraphs
            if line_index == stop.line_index:
                paragraphs = paragraphs[:stop.paragraph_index + 1]
            if line_index == start.line_index:
                paragraphs = paragraphs[start.paragraph_index:]
            for p_index, paragraph in enumerate(paragraphs):
                if p_index == stop.paragraph_index:
                    paragraph = paragraph[:stop.char_index]
                if p_index == start.paragraph_index:
                    paragraph = paragraph[start.char_index:]
                yield paragraph

    def _paste_from_clipboard(self):
        """Paste text from clipboard."""
        try:
            clipboard_text = pg.scrap.get(SCRAP_TEXT).decode('utf-8')
            if clipboard_text:
                clipboard_text = clipboard_text.split('\n')
                for i, line in enumerate(clipboard_text):
                    if i == 0:
                        self._insert_text(clipboard_text[0])
                    elif i == len(clipboard_text) - 1:
                        self._create_newline()
                        self._insert_text(line)
        except pg.error:
            pass

    def _cut_to_clipboard(self):
        """Cut selected text to clipboard."""
        self._copy_to_clipboard()
        self._delete_selection()

    def draw(self, screen: pg.Surface, render_context: RenderContext):
        # Draw background
        if self.is_focused:
            color = self.clicked_color
        elif self.is_hovered:
            color = self.hover_color
        else:
            color = self.bg_color
        pg.draw.rect(screen, color, self.rect)

        # Draw border
        border_width = self.border_width + 1 if self.is_focused else self.border_width
        pg.draw.rect(screen, self.border_color, self.rect, border_width)

        # Text rendering area with padding
        text_area = pg.Rect(
            self.rect.left + self.padding,
            self.rect.top + self.padding,
            self.rect.width - 2 * self.padding,
            self.rect.height - 2 * self.padding
        )

        # Set clipping region
        clip_rect = screen.get_clip()
        screen.set_clip(text_area)

        # Draw text or placeholder
        y_pos = text_area.top + self.padding
        selection_start, selection_end = self._get_selection_range()
        for line_index, line in enumerate(self.lines):
            for paragraph_index, paragraph in enumerate(line.paragraphs):
                self._draw_selection(
                    line_index, paragraph, paragraph_index, screen, selection_end, selection_start, text_area, y_pos
                )

                text_surface = self.font.render(paragraph, True, self.text_color)
                screen.blit(text_surface, (text_area.left, y_pos))

                self.draw_cursor(screen, line_index, paragraph, paragraph_index, y_pos, text_area)

                y_pos += self.line_height

        # Restore clip rect
        screen.set_clip(clip_rect)

    def _draw_selection(
            self, line_index: int, paragraph: str, paragraph_index: int, screen: pg.Surface, selection_end: Cursor,
            selection_start: Cursor, text_area: pg.Rect, y_pos: int
    ):
        if self.selection_start is None:
            return
        if line_index < selection_start.line_index or line_index > selection_end.line_index:
            return
        if line_index == selection_start.line_index and paragraph_index < selection_start.paragraph_index:
            return
        if line_index == selection_end.line_index and paragraph_index > selection_end.paragraph_index:
            return
        start_char_index = 0
        if selection_start.line_index == line_index and selection_start.paragraph_index == paragraph_index:
            start_char_index = selection_start.char_index

        end_char_index = len(paragraph)
        if selection_end.line_index == line_index and selection_end.paragraph_index == paragraph_index:
            end_char_index = selection_end.char_index

        highlight_offset = self.font.size(paragraph[:start_char_index])[0]
        highlight_width = self.font.size(paragraph[start_char_index:end_char_index])[0]
        selection_rect = pg.Rect(
            text_area.left + highlight_offset, y_pos,
            highlight_width, self.line_height
        )
        pg.draw.rect(screen, (100, 150, 200), selection_rect)

    def draw_cursor(self, screen: pg.Surface, line_index: int, paragraph: str, paragraph_index: int, y_pos: int, text_area: pg.Rect):
        if self.is_focused:
            cursor_visible = (pg.time.get_ticks() // 500) % 2 == 0
            if cursor_visible:
                if self.cursor.line_index == line_index and self.cursor.paragraph_index == paragraph_index:
                    x_pos = text_area.left + self.font.size(paragraph[:self.cursor.char_index])[0]
                    pg.draw.line(screen, self.text_color, (x_pos, y_pos), (x_pos, y_pos + self.line_height), 2)



from abc import ABC, abstractmethod
from typing import final

import pygame as pg

from peng_ui.utils import RenderContext


class BaseElement(ABC):
    """
    Defines a base element class with different methods to implement functionality of ui elements.
    """
    def __init__(self, rect: pg.Rect):
        self.visible: bool = True
        self.is_hovered: bool = False
        self.rect: pg.Rect = rect
        self.is_clicked: bool = False

    @abstractmethod
    def handle_event(self, event: pg.event.Event):
        """
        Handle an event.
        :param event: The pygame event to handle
        """
        if event.type == pg.MOUSEMOTION:
            self.is_hovered = self.rect.collidepoint(event.pos)
        elif event.type == pg.MOUSEBUTTONUP:
            if self.is_hovered:
                self.is_clicked = True

    @abstractmethod
    def draw(self, screen: pg.Surface, render_context: RenderContext):
        """
        Draw the element to the given screen surface. Do not call this directly. Prefer render().

        :param screen: The screen surface to draw on.
        :param render_context: The render context to use for rendering.
        """
        pass

    def finalize(self):
        """
        Called at the end of a frame.
        """
        self.is_clicked = False

    @final
    def render(self, screen: pg.Surface, render_context: RenderContext):
        """
        Draw the element to the given screen surface.

        :param screen: The screen surface to draw on.
        :param render_context: The render context to use for rendering.
        """
        if self.visible:
            self.draw(screen, render_context)
        self.finalize()



from typing import Optional, Tuple

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.utils import RenderContext, ColorType, load_font

SCRAP_TEXT = 'text/plain;charset=utf-8'


class EditField(BaseElement):
    def __init__(
            self, rect: pg.Rect, text: str = "", placeholder: str = "",
            bg_color: ColorType = (40, 40, 40), hover_color: ColorType = (60, 60, 60),
            clicked_color: ColorType = (90, 90, 100), border_color: ColorType = (140, 140, 140),
            text_color: ColorType = (200, 200, 200)
    ):
        super().__init__(rect)
        self.text = text
        self.placeholder = placeholder

        self.bg_color = bg_color
        self.hover_color = hover_color
        self.clicked_color = clicked_color
        self.border_color = border_color
        self.text_color = text_color
        self.border_width = 2

        self.cursor_pos: int = len(text)  # Cursor position in text
        self.selection_start: Optional[int] = None  # Start of text selection
        self.is_focused: bool = False  # Whether the field is focused
        self.mouse_down_pos: Optional[int] = None  # For tracking drag selection
        self.text_offset: int = 0  # Horizontal scroll offset for text

        self.font = load_font()

    def _get_char_index_at_pos(self, pos: Tuple[int, int]) -> int:
        """Calculate the character index in text based on mouse position."""
        # Convert screen position to relative position within edit field
        rel_x = pos[0] - self.rect.x - 5 + self.text_offset  # Add offset to account for scrolling

        if not self.text:
            return 0

        # If click is before text start
        if rel_x <= 0:
            return 0

        # Find the closest character position
        for i in range(len(self.text) + 1):
            text_width = self.font.size(self.text[:i])[0]
            if rel_x <= text_width:
                # Check if we're closer to previous or current position
                if i > 0:
                    prev_width = self.font.size(self.text[:i - 1])[0]
                    if rel_x - prev_width < text_width - rel_x:
                        return i - 1
                return i

        return len(self.text)

    def handle_event(self, event: pg.event.Event):
        super().handle_event(event)

        # Handle mouse button down - start selection
        if event.type == pg.MOUSEBUTTONDOWN and event.button == 1:
            if self.is_hovered:
                self.is_focused = True
                # Calculate cursor position from mouse click
                self.cursor_pos = self._get_char_index_at_pos(event.pos)
                self.selection_start = self.cursor_pos
                self.mouse_down_pos = self.cursor_pos
            else:
                self.is_focused = False
                self.selection_start = None

        # Handle mouse drag - update selection
        if event.type == pg.MOUSEMOTION:
            if self.mouse_down_pos is not None and pg.mouse.get_pressed()[0]:
                if self.is_hovered or self.is_focused:
                    self.cursor_pos = self._get_char_index_at_pos(event.pos)

        # Handle mouse button up - finish selection
        if event.type == pg.MOUSEBUTTONUP and event.button == 1:
            if self.mouse_down_pos is not None:
                # If no text was selected (click without drag), clear selection
                if self.selection_start == self.cursor_pos:
                    self.selection_start = None
            self.mouse_down_pos = None

        # Handle keyboard input only if focused
        if self.is_focused:
            if event.type == pg.KEYDOWN:
                ctrl_pressed = event.mod & pg.KMOD_CTRL
                shift_pressed = event.mod & pg.KMOD_SHIFT

                # Handle text input
                if event.key == pg.K_BACKSPACE:
                    self._handle_backspace(ctrl_pressed)
                elif event.key == pg.K_DELETE:
                    self._handle_delete(ctrl_pressed)
                elif event.key == pg.K_LEFT:
                    if ctrl_pressed:
                        self._move_cursor_word_left(shift_pressed)
                    else:
                        self._move_cursor_left(shift_pressed)
                elif event.key == pg.K_RIGHT:
                    if ctrl_pressed:
                        self._move_cursor_word_right(shift_pressed)
                    else:
                        self._move_cursor_right(shift_pressed)
                elif event.key == pg.K_HOME:
                    self._move_cursor_home(shift_pressed)
                elif event.key == pg.K_END:
                    self._move_cursor_end(shift_pressed)
                elif event.key == pg.K_a and ctrl_pressed:
                    self._select_all()
                elif event.key == pg.K_c and ctrl_pressed:
                    self._copy_to_clipboard()
                elif event.key == pg.K_v and ctrl_pressed:
                    self._paste_from_clipboard()
                elif event.key == pg.K_x and ctrl_pressed:
                    self._cut_to_clipboard()
                elif event.unicode and event.unicode.isprintable():
                    self._insert_text(event.unicode)

    def _find_word_start(self, pos: int) -> int:
        """Find the start position of the word at or before the given position."""
        if pos == 0:
            return 0

        # Move back to skip whitespace
        while pos > 0 and self.text[pos - 1].isspace():
            pos -= 1

        # Move back to the start of the previous word
        while pos > 0 and not self.text[pos - 1].isspace():
            pos -= 1

        return pos

    def _find_word_end(self, pos: int) -> int:
        """Find the end position of the word at or after the given position."""
        text_len = len(self.text)

        if pos >= text_len:
            return text_len

        # Move forward to skip whitespace
        while pos < text_len and self.text[pos].isspace():
            pos += 1

        # Move forward to the end of the current word
        while pos < text_len and not self.text[pos].isspace():
            pos += 1

        return pos

    def _update_text_offset(self):
        """Update text offset to keep cursor visible."""
        padding = 5
        visible_width = self.rect.width - 2 * padding

        # Calculate cursor position in text coordinates
        text_before_cursor = self.text[:self.cursor_pos]
        cursor_x = self.font.size(text_before_cursor)[0]

        # Adjust offset to keep cursor visible
        # If cursor is to the right of visible area, scroll right
        if cursor_x - self.text_offset > visible_width:
            self.text_offset = cursor_x - visible_width

        # If cursor is to the left of visible area, scroll left
        if cursor_x < self.text_offset:
            self.text_offset = cursor_x

        # Don't scroll past the beginning
        if self.text_offset < 0:
            self.text_offset = 0

    def _insert_text(self, char: str):
        """Insert a character at the cursor position."""
        self._delete_selection()
        self.text = self.text[:self.cursor_pos] + char + self.text[self.cursor_pos:]
        self.cursor_pos += len(char)
        self.selection_start = None
        self._update_text_offset()

    def _move_cursor_left(self, shift_pressed: bool):
        """Move cursor left, optionally extending selection."""
        if shift_pressed:
            if self.selection_start is None:
                self.selection_start = self.cursor_pos
            if self.cursor_pos > 0:
                self.cursor_pos -= 1
        else:
            if self.selection_start is not None:
                self.cursor_pos = min(self.selection_start, self.cursor_pos)
                self.selection_start = None
            elif self.cursor_pos > 0:
                self.cursor_pos -= 1
        self._update_text_offset()

    def _move_cursor_right(self, shift_pressed: bool):
        """Move cursor right, optionally extending selection."""
        if shift_pressed:
            if self.selection_start is None:
                self.selection_start = self.cursor_pos
            if self.cursor_pos < len(self.text):
                self.cursor_pos += 1
        else:
            if self.selection_start is not None:
                self.cursor_pos = max(self.selection_start, self.cursor_pos)
                self.selection_start = None
            elif self.cursor_pos < len(self.text):
                self.cursor_pos += 1
        self._update_text_offset()

    def _move_cursor_home(self, shift_pressed: bool):
        """Move cursor to the start of the text."""
        if shift_pressed and self.selection_start is None:
            self.selection_start = self.cursor_pos
        self.cursor_pos = 0
        if not shift_pressed:
            self.selection_start = None
        self._update_text_offset()

    def _move_cursor_end(self, shift_pressed: bool):
        """Move cursor to the end of the text."""
        if shift_pressed and self.selection_start is None:
            self.selection_start = self.cursor_pos
        self.cursor_pos = len(self.text)
        if not shift_pressed:
            self.selection_start = None
        self._update_text_offset()

    def _move_cursor_word_left(self, shift_pressed: bool):
        """Move cursor to the start of the previous word."""
        if shift_pressed and self.selection_start is None:
            self.selection_start = self.cursor_pos

        self.cursor_pos = self._find_word_start(self.cursor_pos)

        if not shift_pressed:
            self.selection_start = None
        self._update_text_offset()

    def _move_cursor_word_right(self, shift_pressed: bool):
        """Move cursor to the end of the current/next word."""
        if shift_pressed and self.selection_start is None:
            self.selection_start = self.cursor_pos

        self.cursor_pos = self._find_word_end(self.cursor_pos)

        if not shift_pressed:
            self.selection_start = None
        self._update_text_offset()

    def _handle_backspace(self, ctrl_pressed: bool = False):
        """Handle backspace key."""
        if not self._delete_selection():
            if ctrl_pressed:
                # Delete word to the left
                new_pos = self._find_word_start(self.cursor_pos)
                if new_pos < self.cursor_pos:
                    self.text = self.text[:new_pos] + self.text[self.cursor_pos:]
                    self.cursor_pos = new_pos
            elif self.cursor_pos > 0:
                self.text = self.text[:self.cursor_pos - 1] + self.text[self.cursor_pos:]
                self.cursor_pos -= 1
        self._update_text_offset()

    def _handle_delete(self, ctrl_pressed: bool = False):
        """Handle delete key."""
        if not self._delete_selection():
            if ctrl_pressed:
                # Delete word to the right
                new_pos = self._find_word_end(self.cursor_pos)
                if new_pos > self.cursor_pos:
                    self.text = self.text[:self.cursor_pos] + self.text[new_pos:]
            elif self.cursor_pos < len(self.text):
                self.text = self.text[:self.cursor_pos] + self.text[self.cursor_pos + 1:]
        self._update_text_offset()

    def _get_selection_range(self) -> tuple[int, int]:
        """Get the start and end of the current selection (ordered)."""
        if self.selection_start is None:
            return self.cursor_pos, self.cursor_pos
        return min(self.selection_start, self.cursor_pos), max(self.selection_start, self.cursor_pos)

    def _delete_selection(self) -> bool:
        """Delete selected text if any. Returns True if text was deleted."""
        start, end = self._get_selection_range()
        if start != end:
            self.text = self.text[:start] + self.text[end:]
            self.cursor_pos = start
            self.selection_start = None
            return True
        return False

    def _select_all(self):
        """Select all text."""
        self.selection_start = 0
        self.cursor_pos = len(self.text)

    def _copy_to_clipboard(self):
        """Copy selected text to clipboard."""
        start, end = self._get_selection_range()
        if start != end:
            pg.scrap.put(SCRAP_TEXT, self.text[start:end].encode('utf-8'))

    def _paste_from_clipboard(self):
        """Paste text from clipboard."""
        try:
            clipboard_text = pg.scrap.get(SCRAP_TEXT).decode('utf-8')
            if clipboard_text:
                # Remove newlines and other problematic characters
                clipboard_text = clipboard_text.replace('\n', '').replace('\r', '')
                self._insert_text(clipboard_text)
        except pg.error:
            pass  # Clipboard might not be available

    def _cut_to_clipboard(self):
        """Cut selected text to clipboard."""
        start, end = self._get_selection_range()
        if start != end:
            pg.scrap.put(SCRAP_TEXT, self.text[start:end].encode('utf-8'))
            self._delete_selection()

    def draw(self, screen: pg.Surface, render_context: RenderContext):
        # Draw background
        if self.is_focused:
            color = self.clicked_color
        elif self.is_hovered:
            color = self.hover_color
        else:
            color = self.bg_color
        pg.draw.rect(screen, color, self.rect)

        # Draw border (thicker if focused)
        border_width = self.border_width + 1 if self.is_focused else self.border_width
        pg.draw.rect(screen, self.border_color, self.rect, border_width)

        # Text rendering area with padding
        padding = 5
        text_area = pg.Rect(
            self.rect.left + padding,
            self.rect.top + padding,
            self.rect.width - 2 * padding,
            self.rect.height - 2 * padding
        )

        # Create a subsurface for clipping
        clip_rect = screen.get_clip()
        screen.set_clip(text_area)

        text_x = text_area.left - self.text_offset
        text_y = self.rect.centery

        # Draw selection highlight if there's a selection
        if self.selection_start is not None and self.is_focused:
            start, end = self._get_selection_range()
            if start != end:
                # Calculate selection rectangle
                text_before_selection = self.text[:start]
                selected_text = self.text[start:end]

                before_width = self.font.size(text_before_selection)[0]
                selection_width = self.font.size(selected_text)[0]

                selection_rect = pg.Rect(
                    text_x + before_width,
                    text_area.top,
                    selection_width,
                    text_area.height
                )
                # Draw selection highlight
                pg.draw.rect(screen, (100, 150, 200), selection_rect)

        # Render text or placeholder
        if self.text:
            text_surface = self.font.render(self.text, True, self.text_color)
            text_rect = text_surface.get_rect(midleft=(text_x, text_y))
            screen.blit(text_surface, text_rect)
        elif not self.is_focused and self.placeholder:
            # Draw placeholder text in a dimmer color
            placeholder_color = (100, 100, 100)
            placeholder_surface = self.font.render(self.placeholder, True, placeholder_color)
            placeholder_rect = placeholder_surface.get_rect(midleft=(text_x, text_y))
            screen.blit(placeholder_surface, placeholder_rect)

        # Draw cursor if focused
        if self.is_focused:
            # Blink cursor (using time-based blinking)
            cursor_visible = (pg.time.get_ticks() // 500) % 2 == 0

            if cursor_visible:
                # Calculate cursor position
                text_before_cursor = self.text[:self.cursor_pos]
                cursor_x = text_x + self.font.size(text_before_cursor)[0]

                cursor_height = text_area.height
                cursor_top = text_area.top

                # Draw cursor line
                pg.draw.line(
                    screen,
                    self.text_color,
                    (cursor_x, cursor_top),
                    (cursor_x, cursor_top + cursor_height),
                    2
                )

        # Restore original clip rect
        screen.set_clip(clip_rect)



from typing import Optional

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.utils import RenderContext, ColorType


class Button(BaseElement):
    def __init__(
            self, rect: pg.Rect, text: str = "",
            bg_color: ColorType = (180, 180, 180), hover_color: ColorType = (190, 190, 190),
            clicked_color: ColorType = (200, 200, 210), border_color: ColorType = (100, 100, 100),
            text_color: ColorType = (0, 0, 0)
    ):
        super().__init__(rect)
        self.text = text

        self.bg_color = bg_color
        self.hover_color = hover_color
        self.clicked_color = clicked_color
        self.border_color = border_color
        self.text_color = text_color
        self.border_width = 2

        self.text_surface: Optional[pg.Surface] = None

    def handle_event(self, event: pg.event.Event):
        super().handle_event(event)

    def draw(self, screen: pg.Surface, render_context: RenderContext):
        # Draw button background
        if self.is_hovered:
            color = self.clicked_color if pg.mouse.get_pressed()[0] else self.hover_color
        else:
            color = self.bg_color
        pg.draw.rect(screen, color, self.rect)

        # Draw border
        pg.draw.rect(screen, self.border_color, self.rect, self.border_width)

        # Render text
        if self.text:
            if self.text_surface is None:
                self.text_surface = render_context.font.render(self.text, True, self.text_color)
            text_rect = self.text_surface.get_rect(center=self.rect.center)
            screen.blit(self.text_surface, text_rect)



#!/usr/bin/env python3

import pygame as pg

from peng_ui.elements import TextField, EditField, Label, Button
from peng_ui.viewer import Viewer

class MyViewer(Viewer):
    def __init__(self):
        super().__init__(screen_size=(800, 600))
        self.button = Button(pg.Rect(50, 50, 120, 40), "Click Me!")
        self.label = Label(pg.Rect(50, 100, 120, 40), "This is text :)")
        self.edit_field = EditField(pg.Rect(200, 50, 220, 40), "Edit me...")
        self.text_field = TextField(pg.Rect(200, 100, 520, 440), "This is a long text :).\nIt supports multi-lines, copy-paste, selection, ...")

    def tick(self):
        if self.button.is_clicked:
            self.text_field.set_text(self.edit_field.text)
        if self.label.is_clicked:
            print('label clicked')


if __name__ == '__main__':
    viewer = MyViewer()
    viewer.run()



import warnings
from typing import Union, Tuple

import pygame as pg


ColorType = Union[Tuple[int, int, int], pg.Color]


class RenderContext:
    def __init__(self, font: pg.font.Font):
        self.font = font
        self.mouse_pressed = False

    @staticmethod
    def default():
        font = load_font()
        return RenderContext(font)


def load_font():
    """
    Helper function to load the default font.
    :return: The font to use
    """
    try:
        font = pg.font.Font(None, 36)
    except pg.error:
        warnings.warn("Warning: Could not load default font. Text will not be rendered.")
        font = None
    return font


def clamp(n, minn, maxn):
    return max(min(maxn, n), minn)



from typing import Tuple, Optional, List, Union, Iterable

import pygame as pg

from peng_ui.elements.base_element import BaseElement
from peng_ui.container import Container
from peng_ui.utils import RenderContext


class Viewer:
    def __init__(self, title: str = "Window", screen_size: Tuple[int, int] = (800, 600), flags: int = 0):
        pg.init()
        self.screen = pg.display.set_mode(screen_size, flags)
        pg.scrap.init()
        pg.display.set_caption(title)
        self.running = True
        self.render_context = RenderContext.default()
        self._element_cache: Optional[List[Union[Container, BaseElement]]] = None

    def run(self):
        while self.running:
            self.handle_events()
            self.tick()
            self.render()

        pg.quit()

    def iter_elements(self) -> Iterable[Union[BaseElement, Container]]:
        """
        Iter over all elements in the container.
        :return: Iterable of BaseElement objects.
        """
        if self._element_cache is None:
            self._element_cache = [elem for elem in self.__dict__.values() if isinstance(elem, (BaseElement, Container))]

        yield from self._element_cache

    def handle_events(self):
        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.running = False
                break
            for elem in self.iter_elements():
                elem.handle_event(event)

    def tick(self):
        pass

    def render(self):
        self.screen.fill(0)
        for elem in self.iter_elements():
            elem.render(self.screen, self.render_context)
        pg.display.flip()
